
from vsg.rules import case_rule
from vsg import utils


class rule_010(case_rule):
    '''
    Entity rule 010 checks the "end" keyword has proper case.
    '''

    def __init__(self):
        case_rule.__init__(self, 'entity', '010', 'isEndEntityDeclaration')
        self.solution = 'Change "end" keyword to '

    def _extract(self, oLine):
        return utils.extract_first_keyword(oLine)




from vsg.rules import single_space_before_rule


class rule_003(single_space_before_rule):
    '''
    Case rule 003 checks for a single space before the "is" keyword.
    '''

    def __init__(self):
        single_space_before_rule.__init__(self, 'case', '003', 'isCaseIsKeyword', 'is')
        self.solution = 'Ensure a single space exists before the "is" keyword.'




from vsg.rules import indent_rule


class rule_001(indent_rule):
    '''
    Procedure rule 001 checks the indent of procedures.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'procedure', '001', 'isProcedureKeyword')




from vsg import rule

import re


class rule_008(rule.rule):
    '''
    Whitespace rule 008 checks for spaces after "std_logic_vector"
    '''

    def __init__(self):
        rule.rule.__init__(self, 'whitespace', '008')
        self.phase = 2
        self.solution = 'Remove spaces after "std_logic_vector".'

    def _analyze(self, oFile, oLine, iLineNumber):
        if re.match('^.*std_logic_vector\s+\(', oLine.line, re.IGNORECASE):
            self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub(r'([r|R])\s+\(', r'\1(', oLine.line, 1, re.IGNORECASE))




def indent_of_comments_above(self, oFile, iLineNumber):
    '''
    Checks the indent level of consecutive comment lines above the line number given.

    Parameters:

      self: (rule object)

      oFile: (vhdlFile object)

      iLineNumber: (integer)
    '''
    iIndex = 0
    while iLineNumber - iIndex > 1:
        iIndex += 1
        iPreviousIndex = iLineNumber - iIndex
        if not oFile.lines[iPreviousIndex].isComment:
            break
        else:
            if not oFile.lines[iPreviousIndex].line.index('--') == oFile.lines[iLineNumber].indentLevel * self.indentSize:
                self.add_violation(iPreviousIndex)
                self.dFix['violations'][iPreviousIndex] = oFile.lines[iLineNumber].indentLevel
            else:
                oFile.lines[iPreviousIndex].indentLevel = oFile.lines[iLineNumber].indentLevel



import re


def function(dVars, oLine):
    '''
    Classifies function statements.

    [pure|impure] function identifier [ (parameter_interface_list) ] return type_mark is
      [ subprogram_declarative_part ]
    begin
      { sequential_statement }
    end [function] [identifier] ;

    Sets the following line attributes:

      * insideFunction
      * insideFunctionDeclarative
      * isFunctionParameter
      * isFunctionParameterEnd
      * isFunctionBegin
      * isFunctionKeyword
      * isFunctionEnd
      * isFunctionReturn
      * isFunctionIs

    Modifies the following variables:

      * iOpenParenthesis
      * iCurrentIndentLevel
      * fFunctionParameterEndDetected
      * fFunctionIsDetected
      * fFunctionBeginDetected
    '''
    _classify_function_keyword(dVars, oLine)
    if oLine.insideFunction:
        dVars['iOpenParenthesis'] += oLine.lineNoComment.count('(')
        dVars['iCloseParenthesis'] += oLine.lineNoComment.count(')')
        _classify_begin_keyword(dVars, oLine)
        _classify_return_keyword(dVars, oLine)
        _classify_parameter(dVars, oLine)
        _classify_parameter_end(dVars, oLine)
        _classify_end_keyword(dVars, oLine)


def _classify_function_keyword(dVars, oLine):
    if re.match('^\s*function\s', oLine.lineLower) or re.match('^\s*impure\s', oLine.lineLower) or re.match('^\s*pure\s', oLine.lineLower):
        oLine.isFunctionKeyword = True
        oLine.insideFunction = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1


def _classify_begin_keyword(dVars, oLine):
    if re.match('^\s*begin', oLine.lineLower):
        oLine.isFunctionBegin = True
        oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
        dVars['fFunctionBeginDetected'] = True


def _classify_return_keyword(dVars, oLine):
    if re.match('^\s*return', oLine.lineLower) and dVars['fFunctionBeginDetected']:
        oLine.isFunctionReturn = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']


def _classify_end_keyword(dVars, oLine):
    if oLine.insidePackage:
        if re.match('^.*return\s+\w+;', oLine.line, re.IGNORECASE):
            oLine.isFunctionEnd = True
            oLine.isFunctionReturn = True
            dVars['iCurrentIndentLevel'] -= 1
            dVars['fFunctionBeginDetected'] = False
            dVars['fFunctionParameterEndDetected'] = False
    else:
        if re.match('^\s*end', oLine.lineLower) and not oLine.isEndIfKeyword and \
           not oLine.isEndCaseKeyword and not oLine.isForLoopEnd and not oLine.isWhileLoopEnd and dVars['fFunctionBeginDetected']:
            oLine.isFunctionEnd = True
            oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
            dVars['iCurrentIndentLevel'] -= 1
            dVars['fFunctionBeginDetected'] = False
            dVars['fFunctionParameterEndDetected'] = False


def _classify_parameter(dVars, oLine):
    if re.match('^.*\w+\s*:\s*\w+', oLine.line) and not dVars['fFunctionParameterEndDetected']:
        oLine.isFunctionParameter = True
        if not oLine.indentLevel:
            oLine.indentLevel = dVars['iCurrentIndentLevel']


def _classify_parameter_end(dVars, oLine):
    if ')' in oLine.line and dVars['iOpenParenthesis'] == dVars['iCloseParenthesis']:
        dVars['iOpenParenthesis'] = 0
        dVars['iCloseParenthesis'] = 0
        oLine.isFunctionParameterEnd = True
        dVars['fFunctionParameterEndDetected'] = True




from vsg.rules import prefix_rule
from vsg import utils


class rule_012(prefix_rule):
    '''
    Variable rule 012 checks for prefixes in variable identifiers.
    '''

    def __init__(self):
        prefix_rule.__init__(self, 'variable', '012', 'isVariable')
        self.prefixes = ['v_']
        self.solution = 'Variable'

    def _extract(self, oLine):
        return utils.extract_class_identifier_list(oLine)




def keyword_alignment(self, iLineNumber, sKeyword, lGroup):
    '''
    Checks keywords in a group of line objects are aligned in the same column.

    Parameters:

      self: (rule object)

      iLineNumber: (integer)

      sKeyword: (string)

      lGroup: (list of line objects)
    '''
    iKeywordAlignment = None
    iMaximumKeywordColumn = 0
    sViolationRange = str(iLineNumber) + '-' + str(iLineNumber + len(lGroup) - 1)
    try:
        self.dFix['violations'][sViolationRange]['line'] = {}
    except KeyError:
        self.dFix['violations'][sViolationRange] = {}
        self.dFix['violations'][sViolationRange]['line'] = {}

    for iIndex, oGroupLine in enumerate(lGroup):
        if sKeyword in oGroupLine.line:
            self.dFix['violations'][sViolationRange]['line'][iLineNumber + iIndex] = {}
            self.dFix['violations'][sViolationRange]['line'][iLineNumber + iIndex]['keywordColumn'] = oGroupLine.line.find(sKeyword)

            iMaximumKeywordColumn = get_maximum_keyword_column(oGroupLine, sKeyword, iMaximumKeywordColumn)

            iKeywordAlignment = update_keyword_alignment(oGroupLine, sKeyword, iKeywordAlignment)

            if not iKeywordAlignment == oGroupLine.line.find(sKeyword):
                add_range_violation(self, sViolationRange)

    self.dFix['violations'][sViolationRange]['maximumKeywordColumn'] = iMaximumKeywordColumn


def get_maximum_keyword_column(oLine, sKeyword, iMaximumKeywordColumn):
    '''
    Return the highest column count a keyword is found.

    Parameters:

      oLine: (line object)

      sKeyword: (string)

      iMaximumKeywordColumn: (integer)

    Returns: (integer)
    '''
    if oLine.line.find(sKeyword) > iMaximumKeywordColumn:
        return oLine.line.find(sKeyword)
    return iMaximumKeywordColumn


def update_keyword_alignment(oLine, sKeyword, iKeywordAlignment):
    '''
    Returns the column where the keyword was detected.

    Parameters:

      oLine: (line object)

      sKeyword: (string)

      iKeywordAlignment: (integer)

    Returns: (integer)
    '''
    if not iKeywordAlignment:
        return oLine.line.find(sKeyword)
    return iKeywordAlignment


def add_range_violation(self, sViolationRange):
    '''
    Appends a range violation to the violation list if it does not already exist.

    Parameters:

      self: (rule object)

      sViolationRange: (string)
    '''
    if sViolationRange not in self.violations:
        self.add_violation(sViolationRange)




from vsg import rule
from vsg import fix
from vsg import check


class rule_009(rule.rule):
    '''
    Instantiation rule 009 checks the entity name is uppercase in the instantiation declaration line.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '009'
        self.solution = 'Change entity name to all uppercase.'
        self.phase = 6

    def _analyze(self, oFile, oLine, iLineNumber):
        if oLine.isInstantiationDeclaration:
            sName = oLine.line.split(':')[1].lstrip().split()[0]
            check.is_uppercase(self, sName, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            sLine = oLine.line.split(':')[1].split()[0]
            fix.upper_case(oFile.lines[iLineNumber], sLine)




from vsg.rules import line_above_rule


class rule_003(line_above_rule):
    '''
    Architecture rule 003 checks for a blank line above the architecture
    keyword.
    '''

    def __init__(self):
        line_above_rule.__init__(self, 'architecture', '003', 'isArchitectureKeyword')



import re
import copy


def code_tags(dVars, oLine, oLinePrevious):

    if oLinePrevious.hasCodeTag:
        oLine.hasCodeTag = True
        oLine.codeTags = copy.deepcopy(oLinePrevious.codeTags)

    if re.match('^\s*--\s*vsg_off$', oLine.line):
        oLine.hasCodeTag = True
        oLine.codeTags['vsg_off'] = []

    if re.match('^\s*--\s*vsg_on$', oLinePrevious.line):
        oLine.hasCodeTag = False
        del oLine.codeTags['vsg_off']

    if re.match('^\s*--\s*vsg_off\s+\w', oLine.line):
        oLine.hasCodeTag = True
        lRules = oLine.line.split()[2::]
        for sRule in lRules:
            try:
                oLine.codeTags['vsg_off'].append(sRule)
            except KeyError:
                oLine.codeTags['vsg_off'] = [sRule]

    if re.match('^\s*--\s*vsg_on\s+\w', oLinePrevious.line):
        lRules = oLinePrevious.line.split()[2::]
        for sRule in lRules:
            oLine.codeTags['vsg_off'].remove(sRule)
        if len(oLine.codeTags['vsg_off']) == 0:
            oLine.hasCodeTag = False
            del oLine.codeTags['vsg_off']



import platform
import datetime


class failure():
    '''
    This class holds failure type attributes.
    '''

    def __init__(self, sType):
        self.type = sType
        self.text = None

    def add_text(self, sText):
        try:
            self.text.append(sText)
        except AttributeError:
            self.text = [sText]

    def build_junit(self):
        '''
        Return a formatted failure as a list of strings.
        '''
        lJunit = ['    <failure type="' + self.type + '">']
        for sText in self.text:
            lJunit.append('      ' + sText)
        lJunit.append('    </failure>')

        return lJunit


class testcase():
    '''
    This class holds testcase attributes.
    '''

    def __init__(self, sName=None, sTime=None, sClassName=None):
        self.name = sName
        self.time = sTime
        self.classname = sClassName
        self.failures = None

    def add_failure(self, oFailure):
        try:
            self.failures.append(oFailure)
        except AttributeError:
            self.failures = [oFailure]

    def build_junit(self):
        '''
        Return formatted testcase as a list of strings.
        '''

        lJunit = ['  <testcase name="' + self.name + '" time="' + self.time + '" classname="' + self.classname + '">']

        if self.failures:
            for oFailure in self.failures:
                try:
                    lJunit.extend(oFailure.build_junit())
                except TypeError:
                    pass

        lJunit.append('  </testcase>')

        return lJunit


class testsuite():
    '''
    This class holds testsuite attributes.
    '''

    def __init__(self, sName, sTime):
        self.name = sName
        self.time = sTime
        self.testcases = None
        self.timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%I:%M:%S")

    def add_testcase(self, oTestcase):
        try:
            self.testcases.append(oTestcase)
        except AttributeError:
            self.testcases = [oTestcase]

    def build_junit(self):
        '''
        Return a formatted testsuite as a list of strings.
        '''
        sHeader = '<testsuite'
        sHeader += ' errors="0"'
        sHeader += ' hostname="' + platform.uname()[1] + '"'
        iFailures = 0
        for oTestcase in self.testcases:
            if oTestcase.failures:
                iFailures += len(oTestcase.failures)
        sHeader += ' failures="' + str(iFailures) + '"'
        sHeader += ' timestamp="' + self.timestamp + '"'
        sHeader += ' tests="' + str(len(self.testcases)) + '"'
        sHeader += ' time="' + self.time + '"'
        sHeader += ' name="' + self.name + '"'
        sHeader += '>'
        lJunit = [sHeader]
        lJunit.append('  <properties>')
        lJunit.append('  </properties>')
        for oTestcase in self.testcases:
            lJunit.extend(oTestcase.build_junit())
        lJunit.append('  <system-out>')
        lJunit.append('  </system-out>')
        lJunit.append('  <system-err>')
        lJunit.append('  </system-err>')
        lJunit.append('</testsuite>')

        return lJunit


class xmlfile():
    '''
    This class holds testsuites.
    '''

    def __init__(self, sFileName):
        self.filename = sFileName
        self.testsuites = None

    def add_testsuite(self, oTestsuite):
        try:
            self.testsuites.append(oTestsuite)
        except AttributeError:
            self.testsuites = [oTestsuite]

    def build_junit(self):
        '''
        Return a JUnit XML file formatted as a list of strings.
        '''
        lJunit = ['<?xml version="1.0" ?>']
        for oTestsuite in self.testsuites:
            lJunit.extend(oTestsuite.build_junit())

        return lJunit




from vsg import rule
from vsg import utils

import re


class rule_024(rule.rule):
    '''
    If rule 024 checks for code after the "then" keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'if'
        self.identifier = '024'
        self.solution = 'Move code after "then" keyword to the next line.'
        self.phase = 1

    def _analyze(self, oFile, oLine, iLineNumber):
        if oLine.isThenKeyword and re.match('^.*[\s|)]then\s+\w', oLine.line, re.IGNORECASE):
            self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            utils.split_line_after_word(oFile, iLineNumber, 'then')
            oFile.lines[iLineNumber].isEndIfKeyword = False
            oFile.lines[iLineNumber].isElseKeyword = False
            oFile.lines[iLineNumber + 1].isThenKeyword = False
            oFile.lines[iLineNumber + 1].isIfKeyword = False
            oFile.lines[iLineNumber + 1].isElseIfKeyword = False
            oFile.lines[iLineNumber + 1].indentLevel += 1
            oFile.lines[iLineNumber].isLastEndIf = False
            utils.reclassify_line(oFile, iLineNumber)
            oFile.lines[iLineNumber + 1].isFirstIf = False




from vsg import rule
from vsg import fix
from vsg import check


class rule_007(rule.rule):
    '''
    Generic rule 007 checks generic names are uppercase.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'generic'
        self.identifier = '007'
        self.solution = 'Uppercase generic name.'
        self.phase = 6

    def _analyze(self, oFile, oLine, iLineNumber):
        if oLine.isGenericDeclaration and not oLine.isGenericKeyword:
            check.is_uppercase(self, oLine.line.split()[0], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.upper_case(oFile.lines[iLineNumber], oFile.lines[iLineNumber].line.split()[0])




from vsg.rules import single_space_before_character_rule


class rule_003(single_space_before_character_rule):
    '''
    Instantiation rule 003 checks for a single space before the :
    '''

    def __init__(self):
        single_space_before_character_rule.__init__(self, 'instantiation', '003', 'isInstantiationDeclaration', ':')
        self.solution = 'Ensure only one space before the :.'




from vsg.rules import multiple_spaces_after_rule


class rule_003(multiple_spaces_after_rule):
    '''
    File rule 003 checks for a single space after the **attribute** keyword.
    '''

    def __init__(self):
        multiple_spaces_after_rule.__init__(self, 'file', '003', 'isFileKeyword', 'file')
        self.solution = 'Ensure a single space after the "file" keyword.'




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003
from .rule_004 import rule_004
from .rule_005 import rule_005
from .rule_006 import rule_006
from .rule_007 import rule_007



import re


def library(oLine):

    if not oLine.insideEntity and not oLine.insideArchitecture and not oLine.insidePackage and not oLine.insidePackageBody:
        check_library_keyword(oLine)
        check_use_keyword(oLine)


def check_library_keyword(oLine):
    if re.match('^\s*library', oLine.lineLower):
        oLine.isLibrary = True
        oLine.indentLevel = 0


def check_use_keyword(oLine):
    if re.match('^\s*use', oLine.lineLower):
        oLine.isLibraryUse = True
        oLine.indentLevel = 1



import re


def package_body(dVars, oLine):

    # Check package body declarations
    if re.match('^\s*package\s+body', oLine.lineLower) and not oLine.insidePackageBody:
        oLine.isPackageBodyKeyword = True
        oLine.insidePackageBody = True
        oLine.indentLevel = 0
        dVars['iCurrentIndentLevel'] = 1
    if oLine.insidePackageBody:
        if not oLine.insideProcess and not oLine.insideCase and not oLine.insideComponent and not oLine.insideFunction and not oLine.insideProcedure:
            if re.match('^\s*end\s+', oLine.lineLower):
                oLine.isPackageBodyEnd = True
                oLine.indentLevel = 0
                dVars['iCurrentIndentLevel'] = 0




import re

from vsg import rule


class rule_001(rule.rule):
    '''
    After rule 001 checks for the "after" keyword in signal assignments in clock processes.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'after', '001')
        self.phase = 1
        self.disable = True
        self.magnitude = 1
        self.units = 'ns'
        self.configuration.extend(['magnitude', 'units'])
        self.solution = 'Add after ' + str(self.magnitude) + ' ' + self.units + ' to signal in clock process'

    def _pre_analyze(self):
        self.sequentialStatement = ""

    def _analyze(self, oFile, oLine, iLineNumber):
        if oLine.insideClockProcess:
            if oLine.isSequentialEnd:
                self.sequentialStatement += oLine.lineNoComment.replace('--', '  ')
                if not re.match('^\s*.*\safter\s', self.sequentialStatement):
                    self.add_violation(iLineNumber)
                self.sequentialStatement = ""
            elif oLine.insideSequential:
                self.sequentialStatement += oLine.lineNoComment.replace('--', '  ')
            elif oLine.isSequential:
                self.sequentialStatement = oLine.lineNoComment.replace('--', '  ')

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            sLine = oLine.line
            sNewLine = sLine.replace(';', ' ' + ' '.join(['after', str(self.magnitude), self.units]) + ';')
            oLine.update_line(sNewLine)




from vsg import rule


class rule_015(rule.rule):
    '''Generate rule 015 checks for the generate label and generate keyword on the same line.'''

    def __init__(self):
        rule.rule.__init__(self, 'generate', '015')
        self.solution = 'Merge lines with generate label with generate keyword.'
        self.phase = 1
        self.fixable = False

    def _analyze(self, oFile, oLine, iLineNumber):
        if oLine.isGenerateLabel and not oLine.isGenerateKeyword:
            self.add_violation(iLineNumber)
